import 'dart:ffi';
import 'dart:math';
import "package:collection/collection.dart";
import 'package:verttice_logic/verttice_logic.dart';
import 'package:verttice_ui/src/const/box_sizes.dart';
import 'package:verttice_ui/src/state/verttice_ui_state.dart';
import 'package:verttice_ui/src/widgets/data_factory/common/visual_data_processor.dart';
import 'forms_factory_index.dart';
import "package:flutter/material.dart";
import 'package:flutter/services.dart';
import 'package:provider/provider.dart';
import 'package:after_layout/after_layout.dart';
import 'package:verttice_ui/src/widgets/themed_widgets/themed_widgets_index.dart';

class FormFactory extends StatefulWidget {
  final Map<String, dynamic>? formState;
  final List<FormItem> formItems;
  final bool gridView;
  final int gridViewItemWidth;
  final bool isFilter;
  final bool returnFormStateOnLoad;
  final Function(String name, dynamic _value) onChanged;

  const FormFactory({
    Key? key,
    required this.formState,
    required this.formItems,
    required this.onChanged,
    this.gridView = false,
    this.gridViewItemWidth = 350,
    this.isFilter = false,
    this.returnFormStateOnLoad = false,
  }) : super(key: key);

  @override
  FormFactoryState createState() => FormFactoryState();
}

class FormFactoryState extends State<FormFactory>
    with AfterLayoutMixin<FormFactory> {
  Map<String, dynamic> _formState = {};
  Map<String, FormFieldValidator> _formFieldState = {};

  @override
  void initState() {
    super.initState();
    _formState = DataClassUtils.dynamicToMapStringDynamic(widget.formState);
  }

  @override
  void afterFirstLayout(BuildContext context) {
    if (widget.returnFormStateOnLoad) {
      for (var formItem in widget.formItems) {
        if (formItem.field.defaultValue != null) {
          onChanged(
            formItem.field.name,
            _formState[formItem.field.name] ?? formItem.field.defaultValue,
          );
        }
      }
    }
  }

  @override
  void didUpdateWidget(covariant FormFactory oldWidget) {
    super.didUpdateWidget(oldWidget);

    if (!const MapEquality().equals(_formState, widget.formState)) {
      _formState = DataClassUtils.dynamicToMapStringDynamic(widget.formState);
    }
  }

  onChanged(String name, dynamic value) {
    widget.onChanged(name, value);
    _formState[name] = value;
  }

  bool isValid() {
    bool valid = true;
    for (var element in _formFieldState.entries) {
      if (element.value(_formState[element.key]) != null) {
        valid = false;
        break;
      }
    }
    return valid;
  }

  @override
  Widget build(BuildContext context) {
    return LayoutBuilder(builder: (context, constraints) {
      _formFieldState = {};
      final fields = <Widget>[];
      for (var element in widget.formItems) {
        if (element.field.showOnForm) {
          final validator = element.field.validator;
          final name = element.field.name;
          _formFieldState[name] = (dynamic data) {
            return validator?.valid(data);
          };
          fields.add(
            FormField(
              autovalidateMode: AutovalidateMode.onUserInteraction,
              validator: _formFieldState[name],
              builder: (FormFieldState<dynamic> state) {
                return SizedBox(
                  width: widget.gridView
                      ? widget.gridViewItemWidth.toDouble()
                      : constraints.maxWidth - 14,
                  height: 130,
                  child: Column(
                    crossAxisAlignment: CrossAxisAlignment.start,
                    children: [
                      Expanded(
                        child: _FormItem(
                          columns: widget.formItems,
                          formItem: element,
                          onChanged: onChanged,
                          formState: _formState,
                          isFilter: widget.isFilter,
                          state: state,
                        ),
                      ),
                      Padding(
                        padding: const EdgeInsets.symmetric(
                          vertical: 1.5,
                          horizontal: 4,
                        ),
                        child: SelectableText(
                          _formFieldState[name] != null
                              ? _formFieldState[name]!(_formState[name]) ?? ""
                              : "",
                          maxLines: 1,
                          style: const TextStyle(
                            color: Colors.red,
                            fontSize: 12,
                            overflow: TextOverflow.fade,
                          ),
                        ),
                      )
                    ],
                  ),
                );
              },
            ),
          );
        }
      }
      final columnsAmount = constraints.maxWidth ~/ 375;

      return SizedBox(
        width: constraints.maxWidth,
        child: Consumer<VertticeUIState>(builder: (context, value, _child) {
          if (value.formInReadonlyView && !widget.isFilter) {
            return SizedBox(
              width: constraints.maxWidth,
              height: (widget.formItems.length / columnsAmount) * 15,
              child: Padding(
                padding: const EdgeInsets.all(10),
                child: Wrap(
                  direction: Axis.vertical,
                  children: widget.formItems
                      .map(
                        (e) => Builder(builder: (context) {
                          if (!e.field.showOnForm) {
                            return const SizedBox(
                              height: 0,
                              width: 0,
                            );
                          }
                          final value = visualDataProcessor(
                            e,
                            _formState,
                            context: context,
                          ).value;
                          return Tooltip(
                            message: "${e.field.displayName}: $value",
                            child: Row(
                              children: [
                                SelectableText(
                                  "${e.field.displayName}:",
                                  style: const TextStyle(
                                    color: Colors.blueGrey,
                                    fontWeight: FontWeight.bold,
                                  ),
                                  onTap: () async {
                                    await Clipboard.setData(
                                      ClipboardData(
                                        text: "${e.field.displayName}: $value",
                                      ),
                                    );
                                    ThemedToast.info(
                                      context: context,
                                      message: "Copied to clipboard",
                                    );
                                  },
                                ),
                                sizedBoxW4,
                                SelectableText(
                                  "${value.substring(0, min(30, value.length))}${value.length > 30 ? "..." : ""}",
                                  onTap: () async {
                                    await Clipboard.setData(
                                      ClipboardData(
                                        text: "${e.field.displayName}: $value",
                                      ),
                                    );
                                    ThemedToast.info(
                                      context: context,
                                      message: "Copied to clipboard",
                                    );
                                  },
                                  toolbarOptions: const ToolbarOptions(
                                    copy: true,
                                    selectAll: true,
                                  ),
                                ),
                                sizedBoxW8,
                              ],
                            ),
                          );
                        }),
                      )
                      .toList(),
                ),
              ),
            );
          } else {
            return Padding(
              padding: const EdgeInsets.only(left: 7, right: 7),
              child: Wrap(
                runAlignment: WrapAlignment.spaceAround,
                alignment: WrapAlignment.spaceAround,
                runSpacing: 8,
                spacing: 5,
                direction: widget.gridView ? Axis.horizontal : Axis.vertical,
                children: fields,
              ),
            );
          }
        }),
      );
    });
  }
}

class _FormItem extends StatelessWidget {
  final FormItem formItem;
  final bool isFilter;
  final Map<String, dynamic> formState;
  final Function(String name, dynamic _value) onChanged;
  final FormFieldState<dynamic> state;
  final List<FormItem> columns;
  const _FormItem({
    Key? key,
    required this.formItem,
    required this.onChanged,
    required this.formState,
    required this.state,
    required this.columns,
    this.isFilter = false,
  }) : super(key: key);

  @override
  Widget build(BuildContext context) {
    dynamic _value;
    if (formItem.field.dataBuilder != null) {
      _value = formItem.field.dataBuilder!(formState);
    } else {
      _value = formState[formItem.field.name];
    }

    //////// filter
    if (isFilter) {
      if (formItem.field is TextFF ||
          formItem.field is AddressOnMapFF ||
          formItem.field is ReadonlyTextFF ||
          formItem.field is TextListFF ||
          formItem.field is EmailListFF ||
          formItem.field is TextBoxFF ||
          formItem.field is PhoneNumberListFF ||
          formItem.field is PhoneNumberFF) {
        return TextFormFactory(
          value: _value,
          displayName: formItem.field.displayName,
          name: formItem.field.name,
          type: TextFieldInputFF.text,
          onChanged: onChanged,
          state: state,
        );
      }
      if (formItem.field is IntegerFF ||
          formItem.field is NumberRangeFF ||
          formItem.field is ReadonlyNumberFF ||
          formItem.field is DoubleFF) {
        return NumberRangeFormFactory(
          value: _value,
          displayName: formItem.field.displayName,
          name: formItem.field.name,
          onChanged: onChanged,
          state: state,
        );
      }
      if (formItem.field is SelectFF) {
        return SelectFormFactory(
          onChanged: onChanged,
          pathToField: formItem.field.pathToField,
          displayName: formItem.field.displayName,
          name: formItem.field.name,
          options: (formItem.field as SelectFF).options,
          value: DataClassUtils.dynamicToString(_value),
          state: state,
        );
      } else if (formItem.field is RadioGroupFF) {
        return SelectFormFactory(
          onChanged: onChanged,
          pathToField: formItem.field.pathToField,
          displayName: formItem.field.displayName,
          name: formItem.field.name,
          options: (formItem.field as RadioGroupFF).options,
          value: DataClassUtils.dynamicToString(_value),
          state: state,
        );
      } else if (formItem.field is ExpandSelectFF) {
        return ExpandSelectFormFactory(
          onChanged: onChanged,
          displayName: formItem.field.displayName,
          name: formItem.field.name,
          optionsID: (formItem.field as ExpandSelectFF).optionsID,
          value: DataClassUtils.dynamicToString(_value),
          state: state,
        );
      } else if (formItem.field is ExpandMultiSelectFF) {
        return ExpandSelectFormFactory(
          onChanged: onChanged,
          displayName: formItem.field.displayName,
          name: formItem.field.name,
          optionsID: (formItem.field as ExpandMultiSelectFF).optionsID,
          state: state,
          value: DataClassUtils.dynamicToString(_value),
        );
      } else if (formItem.field is DateRangeFF ||
          formItem.field is ReadonlyDateFF ||
          formItem.field is ReadonlyDateTimeFF ||
          formItem.field is DateFF ||
          formItem.field is ReadonlyDateRangeFF ||
          formItem.field is DateTimeRangeFF ||
          formItem.field is ReadonlyDateTimeRangeFF) {
        return DateRangeFormFactory(
          onChanged: onChanged,
          displayName: formItem.field.displayName,
          name: formItem.field.name,
          value: _value,
          state: state,
        );
      } else if (formItem.field is TimeFF ||
          formItem.field is ReadonlyTimeFF ||
          formItem.field is TimeRangeFF) {
        return TimeRangeFormFactory(
          onChanged: onChanged,
          displayName: formItem.field.displayName,
          name: formItem.field.name,
          value: _value,
          state: state,
        );
      } else if (formItem.field is ColorSelectorFF) {
        final currentValue = _value ?? formItem.field.defaultValue;
        SelectorColors? colorSelectorValue;
        for (var color in SelectorColors.values) {
          if (color.name == currentValue) {
            colorSelectorValue = color;
          }
        }
        return ColorSelectorFormFactory(
          onChanged: onChanged,
          displayName: formItem.field.displayName,
          name: formItem.field.name,
          value: colorSelectorValue,
          colors: (formItem.field as ColorSelectorFF).colors,
        );
      }

      return TextFormFactory(
        value: _value,
        displayName: formItem.field.displayName,
        name: formItem.field.name,
        state: state,
        type: TextFieldInputFF.text,
        onChanged: onChanged,
      );
    }

    //////// end filter

    if (formItem.field is TextFF) {
      return TextFormFactory(
        value: DataClassUtils.dynamicToString(_value, defultValue: ""),
        displayName: formItem.field.displayName,
        name: formItem.field.name,
        type: TextFieldInputFF.text,
        textFormatter: (formItem.field as TextFF).textFormatter,
        state: state,
        onChanged: onChanged,
      );
    } else if (formItem.field is EmailFF) {
      return TextFormFactory(
        value: DataClassUtils.dynamicToString(_value, defultValue: ""),
        displayName: formItem.field.displayName,
        name: formItem.field.name,
        type: TextFieldInputFF.email,
        onChanged: onChanged,
        state: state,
      );
    } else if (formItem.field is IntegerFF) {
      return TextFormFactory(
        displayName: formItem.field.displayName,
        name: formItem.field.name,
        type: TextFieldInputFF.integer,
        onChanged: onChanged,
        state: state,
        value: DataClassUtils.dynamicToString(_value, defultValue: ""),
      );
    } else if (formItem.field is DoubleFF) {
      return TextFormFactory(
        displayName: formItem.field.displayName,
        name: formItem.field.name,
        type: TextFieldInputFF.double,
        onChanged: onChanged,
        state: state,
        value: DataClassUtils.dynamicToString(_value, defultValue: ""),
      );
    } else if (formItem.field is TextBoxFF) {
      return MultilineTextFormFactory(
        value: DataClassUtils.dynamicToString(_value, defultValue: ""),
        displayName: formItem.field.displayName,
        name: formItem.field.name,
        state: state,
        onChanged: onChanged,
      );
    } else if (formItem.field is SelectFF) {
      return SelectFormFactory(
        onChanged: onChanged,
        pathToField: formItem.field.pathToField,
        displayName: formItem.field.displayName,
        name: formItem.field.name,
        options: (formItem.field as SelectFF).options,
        state: state,
        value: _value,
      );
    } else if (formItem.field is RadioGroupFF) {
      return RadioGroupFormFactory(
        displayName: formItem.field.displayName,
        name: formItem.field.name,
        options: DataClassUtils.dynamicToStringList(
            (formItem.field as RadioGroupFF).options),
        onChanged: onChanged,
        state: state,
        value: DataClassUtils.dynamicToString(_value),
      );
    } else if (formItem.field is ExpandSelectFF) {
      return ExpandSelectFormFactory(
        onChanged: onChanged,
        displayName: formItem.field.displayName,
        name: formItem.field.name,
        state: state,
        optionsID: (formItem.field as ExpandSelectFF).optionsID,
        value: DataClassUtils.dynamicToString(_value),
      );
    } else if (formItem.field is ExpandMultiSelectFF) {
      return ExpandMultiSelectFormFactory(
        onChanged: onChanged,
        displayName: formItem.field.displayName,
        pathToField: formItem.field.pathToField,
        name: formItem.field.name,
        state: state,
        optionsID: (formItem.field as ExpandMultiSelectFF).optionsID,
        value: DataClassUtils.dynamicToStringList(_value),
      );
    } else if (formItem.field is CheckboxFF) {
      return CheckboxFormFactory(
        onChanged: onChanged,
        displayName: formItem.field.displayName,
        name: formItem.field.name,
        value: _value,
        state: state,
      );
    } else if (formItem.field is CustomFF) {
      return (formItem.field as CustomFF).builder(CustomFFData(
        columns: columns,
        displayName: formItem.field.displayName,
        name: formItem.field.name,
        context: context,
        onChanged: onChanged,
        state: state,
        data: formState,
      ));
    } else if (formItem.field is DateRangeFF) {
      return DateRangeFormFactory(
        onChanged: onChanged,
        displayName: formItem.field.displayName,
        name: formItem.field.name,
        state: state,
        value: _value,
      );
    } else if (formItem.field is DateTimeRangeFF) {
      return DateTimeRangeFormFactory(
        onChanged: onChanged,
        displayName: formItem.field.displayName,
        name: formItem.field.name,
        value: _value,
        state: state,
      );
    } else if (formItem.field is TimeRangeFF) {
      return TimeRangeFormFactory(
        onChanged: onChanged,
        displayName: formItem.field.displayName,
        name: formItem.field.name,
        state: state,
        value: _value,
      );
    } else if (formItem.field is NumberRangeFF) {
      return NumberRangeFormFactory(
        onChanged: onChanged,
        displayName: formItem.field.displayName,
        name: formItem.field.name,
        state: state,
        value: _value,
      );
    } else if (formItem.field is AddressOnMapFF) {
      return AddressOnMapFormFactory(
        displayName: formItem.field.displayName,
        name: formItem.field.name,
        state: state,
        value: DataClassUtils.dynamicToMapStringDynamic(_value),
        onChanged: onChanged,
      );
    } else if (formItem.field is ReadonlyDateRangeFF) {
      return ReadonlyFormFactory(
        type: ReadonlyFormType.dateRange,
        displayName: formItem.field.displayName,
        name: formItem.field.name,
        value:
            DataClassUtils.dynamicToMapStringDynamic(_value, defultValue: null),
      );
    } else if (formItem.field is ReadonlyDateTimeRangeFF) {
      return ReadonlyFormFactory(
        type: ReadonlyFormType.dateTimeRange,
        displayName: formItem.field.displayName,
        name: formItem.field.name,
        value:
            DataClassUtils.dynamicToMapStringDynamic(_value, defultValue: null),
      );
    } else if (formItem.field is ReadonlyDateFF) {
      return ReadonlyFormFactory(
        type: ReadonlyFormType.date,
        displayName: formItem.field.displayName,
        name: formItem.field.name,
        value: DataClassUtils.dynamicToInt(_value, defultValue: null),
      );
    } else if (formItem.field is ReadonlyDateTimeFF) {
      return ReadonlyFormFactory(
        type: ReadonlyFormType.dateTime,
        displayName: formItem.field.displayName,
        name: formItem.field.name,
        value: DataClassUtils.dynamicToInt(_value, defultValue: null),
      );
    } else if (formItem.field is ReadonlyTimeFF) {
      return ReadonlyFormFactory(
        type: ReadonlyFormType.time,
        displayName: formItem.field.displayName,
        name: formItem.field.name,
        value: DataClassUtils.dynamicToInt(_value, defultValue: null),
      );
    } else if (formItem.field is ReadonlyTextFF ||
        formItem.field is ReadonlyNumberFF) {
      return ReadonlyFormFactory(
        type: ReadonlyFormType.text,
        displayName: formItem.field.displayName,
        name: formItem.field.name,
        value: DataClassUtils.dynamicToString(_value),
      );
    } else if (formItem.field is TimeFF) {
      return TimeFormFactory(
        displayName: formItem.field.displayName,
        name: formItem.field.name,
        state: state,
        value: DataClassUtils.dynamicToInt(_value, defultValue: null),
        onChanged: onChanged,
      );
    } else if (formItem.field is DateFF) {
      return DateFormFactory(
        displayName: formItem.field.displayName,
        name: formItem.field.name,
        state: state,
        value: DataClassUtils.dynamicToInt(_value, defultValue: null),
        onChanged: onChanged,
      );
    } else if (formItem.field is TextListFF) {
      return TextListFormFactory(
        displayName: formItem.field.displayName,
        name: formItem.field.name,
        onChanged: onChanged,
        state: state,
        value: DataClassUtils.dynamicToStringList(_value),
        type: ListType.text,
      );
    } else if (formItem.field is EmailListFF) {
      return TextListFormFactory(
        displayName: formItem.field.displayName,
        name: formItem.field.name,
        onChanged: onChanged,
        state: state,
        value: DataClassUtils.dynamicToStringList(_value),
        type: ListType.email,
      );
    } else if (formItem.field is PhoneNumberListFF) {
      return PhoneNumberListFormFactory(
        displayName: formItem.field.displayName,
        name: formItem.field.name,
        onChanged: onChanged,
        state: state,
        value: _value is List
            ? _value
                .map((e) => DataClassUtils.dynamicToMapStringDynamic(e))
                .toList()
            : [],
      );
    } else if (formItem.field is PhoneNumberFF) {
      return PhoneNumberFormFactory(
        onChanged: onChanged,
        displayName: formItem.field.displayName,
        name: formItem.field.name,
        state: state,
        value: DataClassUtils.dynamicToMapStringDynamic(_value),
      );
    } else if (formItem.field is RichTextFF) {
      return RichTextFormFactory(
        onChanged: onChanged,
        displayName: formItem.field.displayName,
        name: formItem.field.name,
        state: state,
        value: DataClassUtils.dynamicToString(_value),
      );
    } else if (formItem.field is PasswordFF) {
      return PasswordFormFactory(
        onChanged: onChanged,
        displayName: formItem.field.displayName,
        name: formItem.field.name,
        state: state,
        value: DataClassUtils.dynamicToString(_value, defultValue: ""),
      );
    } else if (formItem.field is ColorSelectorFF) {
      final currentValue =
          _value ?? (formItem.field as ColorSelectorFF).defaultValue;
      SelectorColors? colorSelectorValue;
      for (var color in SelectorColors.values) {
        if (color.name == currentValue) {
          colorSelectorValue = color;
        }
      }
      return ColorSelectorFormFactory(
        onChanged: onChanged,
        displayName: formItem.field.displayName,
        name: formItem.field.name,
        value: colorSelectorValue,
        colors: (formItem.field as ColorSelectorFF).colors,
      );
    } else if (formItem.field is ParagraphHeaderFF) {
      return Column(
        crossAxisAlignment: CrossAxisAlignment.center,
        mainAxisAlignment: MainAxisAlignment.center,
        children: [
          sizedBoxH16,
          const Divider(
            thickness: 0.6,
          ),
          sizedBoxH32,
          SelectableText(
            formItem.field.displayName,
            style: TextStyle(
              fontSize: 19,
              color: Colors.blue[700],
              fontWeight: FontWeight.bold,
            ),
          ),
        ],
      );
    } else if (formItem.field is ParagraphFF) {
      return SizedBox(
        height: 90,
        child: Center(
          child: SelectableText(formItem.field.displayName,
              style: const TextStyle(
                fontSize: 13,
              )),
        ),
      );
    }

    return Container();
  }
}
